#undef __NO_VERSION__

/******************************************************************************
*h*
*   Name:               dp.c
*
*   Description:        This is the source code for the Linux Network OS
*		                Specific Module
*
*h*
******************************************************************************/
#pragma ident   "@(#)$RCSfile: dp.c,v $ $Revision: 1.29 $ $Date: 2001/09/26 14:00:45 $ NSC"

#include "nsmtypes.h"
#include "dp83820.h"
#include "nsm.h"
#include "hsm.h"

static char version[] __initdata =
"dp83820 ethernet driver, version 1.10, Mar 5 2002, linux 2.2.x kernel\n";

/* Default initialisation of load time parameters */
UINT MaxTxDesc=250;
UINT MaxRxDesc=250;
UINT SetAutoNeg=1;
UINT RxBufSize=2048;
UINT MediaSpeed=1000;
UINT NCBit=1;
UINT CacheLineSz=512;
UINT Duplex=1;
USHORT vlan=0;

#ifdef PRIORITY_QUEUES

UINT MaxPrio=8;

#else

UINT MaxPrio=1;

#endif

#if defined(MODULE)

MODULE_DESCRIPTION("NSC DP83820 10/100/1000 Gigabit NIC driver");

/* Load time parameters */
MODULE_PARM(MaxTxDesc, "i");
MODULE_PARM(MaxRxDesc, "i");
MODULE_PARM(RxBufSize, "i");
MODULE_PARM(MediaSpeed, "i");
MODULE_PARM(SetAutoNeg, "i");
MODULE_PARM(NCBit, "i");
MODULE_PARM(Duplex, "i");
MODULE_PARM(vlan, "h");
MODULE_PARM_DESC(vlan, "802.1Q Vlan ID (1-4095)");
#endif

#ifdef LINK_AGGR
void LacpRegister(AdapterContext *);
INT  AggrSendPackets(AdapterContext *, PktStruct *, UCHAR, INT *);
INT  AggrSendPacket(AdapterContext *, PktStruct *, UCHAR, INT *);
UCHAR AdapterOpen (struct device *dev);
void AdapterClose (struct device *dev);
#define GEC_FLAG        0x01
//Added by jaggu
extern	char *format(int);
extern	void stringcpy(char *dest, char *src); 
#endif

/* List of the various instances of the DP83820 */
struct device *DevList = NULL;


/******************************************************************************
*f*
*   Name:           DrvProbe
*
*   Description     This routine enumerates Drv ethernet devices on the PCI 
*   				bus, and registers each device found.
*
*   Parameters:     dev: Pointer to the device structure 
*
*   Return Value:   OK or -ENODEV 
*
*f*
******************************************************************************/

int
DrvProbe (struct device *dev)
{
	int 			DevCount=0;
	struct pci_dev 	*pPciDev=NULL;
	UINT  			iobase;
	UINT  			membase;
	AdapterContext 	*pNsc;
	NsmContext 		*pNsm;
	int				version_disp;
	UINT 			cfg_val;
	UINT 			cfg_lat;
	UINT			srr_val;

	/* Check for PCI BIOS being present, return ENODEV if not present */
	if (! pci_present())
	{
#ifdef FAILURE_MESSAGES
		printk("\nPCI bios not present"); 
#endif
	
		/* No such device */
   		return -ENODEV; 
	}

	version_disp = 0;

	/* Find the instances of the device present, and for each instance of the 
	device do the following */

	while ( (pPciDev = pci_find_device (PCI_VENDOR_ID_NS, 
		PCI_DEVICE_ID_83820, pPciDev) ) != NULL)
	{

		/* Get the IOBASE, MEMBASE address  */
		iobase = (UINT) pPciDev->base_address[0];
		membase = (UINT)pPciDev->base_address[1];

		iobase &= PCI_BASE_ADDRESS_IO_MASK;
		membase &= PCI_BASE_ADDRESS_MEM_MASK;

#ifdef IOMAPPED_IO
		
		/* Check if a range of ports is already locked by other drivers */
		if (check_region(iobase,PCI_IO_SIZE))
		{  
#ifdef FAILURE_MESSAGES
			printk ("check_region failed : IO region already in use\n");
#endif
			continue;
		}
#endif		
		/* Allocate space for the device structure, name the device, and also 
		for the private data structure, in this case, the NsmContext. Add the 
		device to the ethernet device list and register the network 
		device. */

		dev = init_etherdev(dev, sizeof(NsmContext));

		if (dev == NULL) 
		{	
#ifdef FAILURE_MESSAGES
			printk (KERN_ERR "%s: Failed to create device struct\n",DRV_NAME);
#endif

			break;
		}

		/* Display version info if Adapter is found */
		if(!version_disp)
		{
			/* Display string only once */
			version_disp = 1;
			printk(version);
		}

#ifdef LINK_AGGR
		//Added by jaggu
		stringcpy(dev->name, format(DevCount));
		printk(KERN_INFO "\nDP83820 Gigabit NIC : %s\n",dev->name);
#else
		printk(KERN_INFO "\nDP83820 Gigabit NIC : %s\n",dev->name);
#endif

		/* Get the pointer to the NsmContext */
		pNsm = (NsmContext *)dev->priv;
		
		pNsm->dev = dev;

		/* Get the pointer to the adapter context */
		pNsc = &pNsm->AdpCtxt ;
		
		/* Allocate the structure to hold the enet_statistics  */
		/* This is reqd here because an ifconfig done can request the statistcs 
			even before an open is done */
		pNsm->pAdapterStats = (struct enet_statistics *) 
			NsmMalloc(sizeof(struct enet_statistics));

		/* Check if the allocation was successful */
		if(pNsm->pAdapterStats == NULL)
		{
#ifdef FAILURE_MESSAGES
			printk(KERN_INFO "Allocation of Statistics struct failed\n");
#endif
			unregister_netdev(dev);
			kfree(dev);
			break;
		}

		pNsc->pNsmContext = (void *)pNsm;

		/* initialize the device data like iobase, irq, and the entry points 
		for the driver in the device structure */

		pNsc->TxMaxDMASize= CacheLineSz;
		pNsc->RxMaxDMASize= CacheLineSz;

		pNsc->MaxPktSize = RxBufSize;

		pci_set_master(pPciDev);
#ifdef MEMMAPPED_IO

		pci_read_config_dword(pPciDev, CFGROM, &cfg_lat);
		if (cfg_lat & 1)//ROMEN issue
			pci_write_config_dword(pPciDev, CFGROM, cfg_lat&~1);	

		dev->base_addr = (ULONG) membase; /* Memory mapped IO */
		pNsc->RegAddr = (UCHAR *)ioremap(membase, PCI_MEM_SIZE);

#elif IOMAPPED_IO

		dev->base_addr = (ULONG) iobase;
		pNsc->RegAddr =  (UCHAR *)dev->base_addr;
#endif

		dev->irq       = pPciDev->irq;
		dev->open      = DrvOpen;
    	dev->stop      = DrvClose;
    	dev->get_stats = DrvGetStatistics;
		dev->do_ioctl  = DrvIoctl;
		dev->hard_start_xmit = DrvSend;
		dev->set_multicast_list = DrvSetMulticastList;
		dev->set_mac_address = DrvSetMacAddress;

		/* Change the mtu function in the device structure as ether_setup sets 
		it to a function which can accepts MTUs of size less than 1500, so 
		change it to a function which can change the MTU to any value 
		less than 65280 and greater than 68. */
		
		dev->change_mtu = ChangeMtu; 

#ifdef IOMAPPED_IO
		
		/* reserve IO region, this locks the ports for later use */
		request_region (iobase, PCI_IO_SIZE, dev->name);

#endif

		/* Work Around to fix 8byte replicating issue */
		NsmRegRead32(pNsm, (pNsc->RegAddr + CFG), &cfg_val);
		NsmRegRead32(pNsm, (pNsc->RegAddr + SRR), &srr_val);

		if(!(cfg_val & PCI64_DET) && (srr_val <= 0x103)) {
			pci_read_config_dword(pPciDev, CFGLAT, &cfg_lat);
			cfg_lat &= 0xFFFF00FF;
           		cfg_lat |= 0xFE00;
			pci_write_config_dword(pPciDev, CFGLAT, cfg_lat);	
		}

		HsmGetMacAddress(pNsc);
		
		/* Install interrupt vector and register the interrupt service 
		routine */
		if((AdapterOpen(dev)) != OK)
		{
		#ifdef FAILURE_MESSAGES
			printk (KERN_ERR "%s: AdapterOpen Failed \n", DRV_NAME);
		#endif
			kfree(pNsm->pAdapterStats);
			unregister_netdev(dev);
			kfree(dev);
			break;
		}

		/* Chain the device */
		pNsm->next = (void *)DevList;
   		DevList = dev;

		/* Update Counters for the number of devices found */
		DevCount++;
		
	#ifdef	LINK_AGGR	
	
		pNsc->FuncStruct.HsmSendPkt = HsmSendPacket;
		pNsc->FuncStruct.HsmSendPkts = HsmSendPkts;
		pNsc->FuncStruct.HsmMulticastAdd = HsmMulticastAdd;
		pNsc->FuncStruct.DrvSend = NULL;
		pNsc->FuncStruct.HsmMulticastDelete = HsmMultiCastDelete;
		pNsc->FuncStruct.HsmSetMacAddress = HsmSetMacAddress;
		pNsc->FuncStruct.HsmRxFilter = HsmRxFilter;
		pNsc->FuncStruct.HsmSetPhyCapabilities = HsmSetPhyCapabilities;
		pNsc->FuncStruct.NsmSetPktStatus = NULL;
		pNsc->FuncStruct.NsmIndicateLinkStatus = NsmIndicateLinkStatus;

		LacpRegister(pNsc);

	#endif
		dev=NULL;
  	}

	if(DevCount == 0)
	{
#ifdef FAILURE_MESSAGES
		printk (KERN_ERR "No device of NSC present\n");
#endif
		return -ENODEV;
	}

	return 0;
}

/******************************************************************************
*f*
*   Name:           DrvOpen
*
*   Description:    DrvOpen opens and initializes a device. This is 
*    				typically called when an ifconfig is done for this 
*					interface.
*
*   Parameters:     dev: Pointer to the device structure 
*
*   Return Value:   OK or -EAGAIN 
*
*f*
******************************************************************************/

STATIC int
DrvOpen (struct device *dev)
{

#if	0	
	UINT 			stat;
	NsmContext 		*pNsm;
	AdapterContext 	*pNsc;
	UINT			i;

	/* Get the pointer to the allocated private data structure, NsmContext, 
	from the device structure */
	pNsm = (NsmContext *)dev->priv;

	/* Get the pointer to the AdapterContext structure from the NsmContext */
	pNsc = &pNsm->AdpCtxt;

	for(i=0; i< MAX_PRI_QUEUE; i++)
	{
		pNsc->RxQueueSz[i] = MaxRxDesc;
		pNsc->TxQueueSz[i] = MaxTxDesc;
	}
	pNsc->TxDrth = TxDrth;		
	pNsc->RxDrth = RxDrth;
	pNsc->TxFlth = TxFlth;
	pNsc->PauseCounterVal = PauseCounterVal;
	pNsc->RxFFLO = 2;
	pNsc->RxFFHI = 8;
	pNsc->RxSTLO = 2;
	pNsc->RxSTHI = 8;
	
	/* Initialize fields which would indicate the capabilities of the hardware,
	 such as the, checksum offload */
	HsmInitContext(pNsc);

	/* Find out the capabilities of the OS and indicate whether the above 
	capabilties are supported by the OS. */
	
	pNsc->OsPrio = MaxPrio;
	pNsc->PhysCapabilities &= ~(VLAN_TAG_INSERTION_GEN_ON | 
		SET_BIG_ENDIAN_MODE | TX_CHKSUM_OFFLOAD_ON_GEN ); 

	if(SetAutoNeg == 0)
	{
		pNsc->PhysCapabilities &= ~AUTO_NEG_ON;  
		pNsc->UserMediaSpeed = MediaSpeed;
	}

	/* Allocate the Tx and Rx lists, based on the number of priority levels 
	supported. Call the HsmInitialise routine of the HSM, this will also setup 
	transmit & receive control. This will set up descriptor lists for each 
	level of priorities on both send and receive side. */

	if ( (stat = HsmInitialize(pNsc) ) == FAILURE )
	{
#ifdef FAILURE_MESSAGES
		printk (KERN_ERR "%s: HsmInitialize Failed \n", DRV_NAME);
#endif
		return -EAGAIN;
	}

	/* Now set the Mac address in the device structure */
	NsmCopy((char *)pNsc->PermMacAddr, (char *)dev->dev_addr, 
		MAC_ADDR_LEN); 

	/* Install interrupt vector and register the interrupt service routine */
	if(request_irq (dev->irq, DrvIsr, SA_SHIRQ,
    	dev->name, dev) != OK)
	{
#ifdef FAILURE_MESSAGES
		printk (KERN_ERR "%s: request_irq Failed \n", DRV_NAME);
#endif
		HsmUnInitialize(pNsc);
		return -EAGAIN;
	}

	/* Reset the adapter  */
	if( ( stat =HsmReset(pNsc, 0) ) == FAILURE)
	{
#ifdef FAILURE_MESSAGES
		printk (KERN_ERR "%s: Resetting the adapter, failed \n", 
			DRV_NAME);
#endif
		free_irq (dev->irq, dev);
		HsmUnInitialize(pNsc);
		return -EAGAIN;
	} 

	/* Call HsmOpen call to open the adapter */
	if( ( stat =HsmOpen(pNsc) ) == FAILURE)
	{
#ifdef FAILURE_MESSAGES
		printk (KERN_ERR "%s: Opening the adapter, failed \n", 
			DRV_NAME);
#endif
		free_irq (dev->irq, dev);
		HsmUnInitialize(pNsc);
		return -EAGAIN;
	}
	
	HsmRxFilter(pNsc, (ACCEPT_PERFECT_MATCH_ON | ACCEPT_ALL_BROADCAST_ON));

	/* Mark the start flag in the device structure */
	dev->start = DEV_STARTED;
	
	dev->tbusy = 0;
	dev->interrupt = 0;

	/* Set the transmit queue length in the device structure */
	dev->tx_queue_len = (MaxTxDesc*MaxPrio);
#endif
	AdapterContext 	*pNsc = &((NsmContext *)dev->priv)->AdpCtxt;

	writel(TXE|RXE, (pNsc->RegAddr+CR));
	if(!(((NsmContext *)dev->priv)->status & NSM_OPEN))
		return -ENOMEM;

	
	/* Increment module reference count, MOD_INC_USE_COUNT */
	MOD_INC_USE_COUNT;

    return OK;
}

/******************************************************************************
*f*
*   Name:           DrvClose
*
*   Description:    This closes the device and reclaims the resources
*
*   Parameters:     dev: Pointer to the device structure 
*
*   Return Value:   OK or ERROR  
*
*f*
******************************************************************************/

STATIC int
DrvClose (struct device *dev)
{

#if 0
	NsmContext 		*pNsm;
	AdapterContext 	*pNsc;

	dev->start = 0;
	dev->tbusy = 1;

	/* Get the pointer to the NsmContext and thereby to the AdapterContext, 
	from the device structure */
	pNsm = (NsmContext *)dev->priv;
	pNsc   = &(pNsm->AdpCtxt);

	/* Stop Tx & Rx by calling HsmClose */
	HsmClose(pNsc); 

	/* Uninstall the interrupt vector, and deregister the interrupt service 
	routine */
	free_irq (dev->irq, dev);

	/* Free the Tx and Rx lists by calling HsmUnInitialize */
	HsmUnInitialize(pNsc);  
#endif
	AdapterContext 	*pNsc = &((NsmContext *)dev->priv)->AdpCtxt;

	writel(TXD|RXD, (pNsc->RegAddr+CR));
	if(!(((NsmContext *)dev->priv)->status & NSM_OPEN))
		return -ENOMEM;

	del_timer(&((NsmContext *)dev->priv)->LedTimer);

	/* Decrement module reference count using MOD_DEC_USE_COUNT */
	MOD_DEC_USE_COUNT;

  	return OK;
}

/******************************************************************************
*f*
*   Name:           DrvSend
*
*   Description:    This sends a packet down to the HSM, to be sent on wire
*
*   Parameters:     dev: Pointer to the device structure
*                   skb: Pointer to the sk_buff which contains the data to be 
*                   sent down
*
*   Return Value:   OK or ERROR
*
*f*
******************************************************************************/


STATIC int
DrvSend (struct sk_buff *skb, struct device *dev)
{
	NsmContext 		*pNsm;
	AdapterContext 	*pNsc;
	UINT			retval=OK, Priority;
#ifdef LINK_AGGR
	INT				CableStatus = 0;
#endif


#ifdef SINGLE_PACKET

	PktInfo			Pkt;
	UCHAR			status;

#elif MULTIPLE_PACKETS

	static PktStruct		Pkts;
	UCHAR 			stat;

#endif
	
	/* Get the pointer to the NsmContext and thereby to the AdapterContext, 
	from the device structure */
	pNsm = (NsmContext *)dev->priv;
	pNsc = &(pNsm->AdpCtxt);

	/* Set the transmit start time in the device structure */
	dev->trans_start = jiffies;

	/* Encapsulate the data which needs to be sent on the wire in a Packet 
	struct */

#ifdef SINGLE_PACKET

	CreatePacketInfo( skb, &Pkt);

#elif MULTIPLE_PACKETS

	CreatePacketStruct( skb, &Pkts);

#endif

	/* Get the priority of the packet, which is going to be sent down from the 
	priority field in the skbuff. */

	/* Call the HSM routine to send the packet on the wire with the priority 
	obtained from the priority field of the skbuff. */ 

#ifdef SINGLE_PACKET

#ifdef	LINK_AGGR
	Pkt.pDLHeader = (VOID *)(skb->data); 
#endif

#ifdef PRIORITY_QUEUES	
	Priority = skb->priority;
#else
	Priority = 0;
#endif
#ifdef LINK_AGGR
        if((pNsc->nFlags) & GEC_FLAG) 
		{
	        status = AggrSendPacket(pNsc,&Pkt,Priority, &CableStatus); 

			/* Check if no link - Ravi */
			if (CableStatus == -1) {
				//NsmDbgMsg("\nCable not connected");
				//retval = ERROR;
				//break;
            	status = HsmSendPacket(pNsc, &Pkt,Priority);
			}
		}
        else
		{
			//printk(KERN_INFO "dp.c :Before calling Hsm Send\n");
            status = HsmSendPacket(pNsc, &Pkt,Priority);
		}
#else
		status = HsmSendPacket(pNsc,&Pkt,Priority);
#endif
	
	if(status == FAILURE)
	{
		retval = ERROR;
	}

#elif MULTIPLE_PACKETS

#ifdef	LINK_AGGR
	Pkts.PktArray[0].pDLHeader = 
		(VOID *)(skb->data); 
#endif

	#ifdef PRIORITY_QUEUES	
		Priority = skb->priority;
	#else
		Priority = 0;
	#endif

	#ifdef LINK_AGGR
      if((pNsc->nFlags) & GEC_FLAG) 
		{
	       	stat = AggrSendPackets(pNsc,&Pkts,Priority, &CableStatus); 
			
			/* Check if no link - Ravi */
				if (CableStatus == -1) 
				{
					//NsmDbgMsg("\nCable not connected");
					//retval = ERROR;
					//break;
            		stat = HsmSendPkts(pNsc, &Pkts,Priority,0);
				}
		}
		else
		{
            stat = HsmSendPkts(pNsc, &Pkts,Priority,0);
		}
	#else
			stat = HsmSendPkts(pNsc,&Pkts,Priority);
	#endif
		if(stat == 0)
		{
			/* MKC change it */
			retval = ERROR;
		}
#endif
	
 		return retval;
}

/******************************************************************************
*f*
*   Name:           DrvSetMulticastList 
*
*   Description:    NsmSetMulticastList sets multicast filters & promiscuous 
*                   mode 
*
*   Parameters:     dev: Pointer to the device structure
*
*   Return Value:   None 
*
*f*
******************************************************************************/

STATIC void
DrvSetMulticastList (struct device *dev)
{
	struct dev_mc_list 	*mc_list;
	MultiList 			*pMultiList;
	MultiList			*pMulti;
	MultiList 			*pMultiNext;
	NsmContext			*pNsm = ((NsmContext *)dev->priv);
	UINT				FilterFlags;

	/* Get the pointer to the NsmContext and thereby to the AdapterContext, 
	from the device structure */
	AdapterContext 		*pNsc = &(pNsm->AdpCtxt);

	FilterFlags = (pNsm->FilterFlags | ACCEPT_PERFECT_MATCH_ON | 
		ACCEPT_ALL_BROADCAST_ON);
    
	/* Setup promiscuous mode */ 
	if (dev->flags & IFF_PROMISC)
	{

		if(pNsm->FilterFlags & PROMISCUOUS_ON)
		{
			FilterFlags |= PROMISCUOUS_OFF;
			FilterFlags &= ~PROMISCUOUS_ON;
		}
		else if (pNsm->FilterFlags & PROMISCUOUS_OFF)
		{
			FilterFlags |= PROMISCUOUS_ON;
			FilterFlags &= ~PROMISCUOUS_OFF;
		}
		else /* The first time */
		{
			FilterFlags |= PROMISCUOUS_ON;
		}
			
	}
 
	/* Receiving all multicast packets  */
	if (dev->flags & IFF_ALLMULTI) 
	{

		if(pNsm->FilterFlags & ACCEPT_ALL_MULTICAST_OFF)
		{
			FilterFlags |= ACCEPT_ALL_MULTICAST_ON;
			FilterFlags &= ~ACCEPT_ALL_MULTICAST_OFF;
		}
		else if (pNsm->FilterFlags & ACCEPT_ALL_MULTICAST_ON)
		{
			FilterFlags |= ACCEPT_ALL_MULTICAST_OFF;
			FilterFlags &= ~ACCEPT_ALL_MULTICAST_ON;
		}
		else /* First time */
		{
			FilterFlags |= ACCEPT_ALL_MULTICAST_ON;
		}
	}		

	/* Call the HSM function to set the receive filters */
	if(FilterFlags)
	{
		HsmRxFilter(pNsc, FilterFlags);
		pNsm->FilterFlags = FilterFlags;
	}

	/* Set the list of multicast addresses in the mc_list */
	mc_list = dev->mc_list;
	pMultiList = NULL;

	/* Get the multicast address list from the mc_list and allocate a 
	MultiList structure and put the multicast addresses into the MultiList 
	structure  */

	while(mc_list)
	{
		pMulti = (MultiList *)NsmMalloc(sizeof(MultiList));
		if(pMulti == NULL)
		{
			if(pMultiList)
			{
				pMulti = pMultiList;
				while(pMulti)
				{
					pMultiNext = pMulti->Next;
					kfree(pMulti);
					pMulti = pMultiNext;
				}
			}
			return;
		}
		NsmCopy(mc_list->dmi_addr, 
			pMulti->MulticastAddr,MAX_ADDR_LEN); 
		pMulti->Next = pMultiList;
		pMultiList = pMulti;
		mc_list = mc_list->next;		        	
	}

	/* Call the HSM function to add the list of multicast addresses to the 
	filter */
	HsmMulticastAdd(pNsc, pMultiList, TRUE);	

	/* Free up the list of MultiList structures */
	pMulti = pMultiList;
	while(pMulti)
	{
		pMultiNext = pMulti->Next;
		kfree(pMulti);
		pMulti = pMultiNext;
	}
}

/******************************************************************************
*f*
*   Name:           DrvSetMacAddress
*
*   Description:    This sets a new MAC address
*
*   Parameters:     dev: Pointer to the device structure
*
*                   addr: The new MAC address that needs to be set 
*
*   Return Value:   OK or ERROR 
*
*f*
******************************************************************************/

int DrvSetMacAddress(struct device *dev, void *addr)
{
	AdapterContext *pNsc = &((NsmContext *)dev->priv)->AdpCtxt;

	void *MacAddr = (void *) (((struct sockaddr *)addr)->sa_data);

	NsmCopy(MacAddr, (void *)dev->dev_addr, MAC_ADDR_LEN);

	NsmCopy(MacAddr, (void *)pNsc->CurrMacAddr, MAC_ADDR_LEN);

	HsmSetMacAddress(pNsc, pNsc->CurrMacAddr); 

	return OK;
}

/******************************************************************************
*f*
*   Name:           DrvIsr 
*
*   Description:    This is the interrupt handler of the driver
*
*   Parameters:    	irq: This represents the IRQ value for which this handler 
*                   was registered
*
*                  	dev_id:  This is a pointer to the device structure
*
*                  	regs:    This is a pointer to a snapshot of the processor's 
*                   context, before the processor entered the interrupt code
* 
*   Return Value:   None
*
*f*
***************************************************************************/

STATIC void
DrvIsr (int irq, void *dev_id, struct pt_regs *regs)
{
	/* Get the pointer to the device structure from the dev_id */
	struct device 	*dev = ( struct device *) dev_id;

	/* Get the pointer to the AdapterContext */
	AdapterContext 	*pNsc = &(((struct NsmInfo *)dev->priv)->AdpCtxt);

	/* Set the interrupt flag in the device structure  */
	dev->interrupt = DEV_INTERRUPTED;

	/* Call the HSMs interrupt service routine */
	HsmIsr (pNsc);

	/* Clear the interrupt flag in the device structure */
	dev->interrupt = DEV_SERVICED_INTERRUPT; 
}


/******************************************************************************
*f*
*   Name:           DrvIoctl 
*
*   Description:    This is the ioctl handler of the driver 
*
*   Parameters:     dev: This is a pointer to the device structure
*
*                   rq:  pointer to a copy of the structure passed by the user 
*                   for this command
*
*                   cmd: This indicates the ioctl command
*
*   Return Value:   OK or -EOPNOTSUPP 
*
*f*
***************************************************************************/

int 
DrvIoctl( struct device *dev, struct ifreq *rq, int cmd )
{

	/* Get the data for the ioctl from the ifreq structure */ 
	void 			*data = &(rq->ifr_data);

	/* Get the pointer to the AdapterContext */
	AdapterContext 	*pNsc = &((NsmContext *)(dev->priv))->AdpCtxt;

	/* Depending on the cmd perform the following */
	switch( cmd )
	{

		/* For setting interface address(dev->dev_addr) */
		case SIOCSIFADDR:

			/* Set the MAC address in the device structure and also the 
			AdapterContext, with the one obtained from the data */

			NsmCopy((const void *)data, (void *)dev->dev_addr, MAC_ADDR_LEN);
			NsmCopy((const void *)data, (void *)pNsc->CurrMacAddr, 
				MAC_ADDR_LEN);
			HsmSetMacAddress(pNsc, pNsc->CurrMacAddr); 
			break;	

		/* The ioctl implementation for sockets, defines 16 ioctls cmds as 
		private to the interface, SIOCSDEVPRIVATE, SIOCSDEVPRIVATE+ 1,.. 
		& so on  */

		case SIOCDEVPRIVATE: 
			break;
		default:
			return -EOPNOTSUPP;

	}
	return OK;
}

/******************************************************************************
*f*
*   Name:           DrvGetStatistics 
*
*   Description:    This is the routine used to get the statistical information
*                   regarding the transmission and reception of packets.
*
*   Parameters:     dev: This is a pointer to the device structure
*
*   Return Value:   Pointer to the statistics structure 
*
*f*
***************************************************************************/

struct enet_statistics *
DrvGetStatistics(struct device *dev)
{
	/* Get the pointer to the NsmContext and thereby to the AdapterContext, 
		from the device structure */
	NsmContext 				*pNsm = (NsmContext *)dev->priv;
	AdapterContext 			*pNsc = &(pNsm->AdpCtxt);

	struct enet_statistics 	*pAdpStats = pNsm->pAdapterStats; 

	/* Update the fields in the StatInfo structure (MacStats) in the 
		AdapterContext, by reading the current values from the MIB 
		registers of the adapter */
	if(HsmGetStatistics(pNsc) == FAILURE) 
	{
		return NULL;
	}
	
	/* Get the statistics from the adapter context structure & set the values 
		of the statistics in the enet_statistics structure */

	pAdpStats->rx_packets = pNsc->MacStats.rxOkCount;
	pAdpStats->rx_errors = pNsc->MacStats.rxErrorCount;
	pAdpStats->rx_dropped = pNsc->MacStats.rxDroppedCount;
	pAdpStats->rx_crc_errors = pNsc->MacStats.rxCrcErrorCount;
	pAdpStats->rx_frame_errors = pNsc->MacStats.rxFaeCount;
	pAdpStats->rx_missed_errors = pNsc->MacStats.rxNoBufferCount;
	
	pAdpStats->tx_errors = pNsc->MacStats.txErrorCount;
	pAdpStats->tx_packets = pNsc->MacStats.txOkCount;
	pAdpStats->tx_dropped = pNsc->MacStats.txDroppedCount;
	
	pAdpStats->tx_carrier_errors = pNsc->MacStats.txCrsLostCount;
	pAdpStats->tx_aborted_errors = pNsc->MacStats.txAbortCount;
	pAdpStats->tx_fifo_errors = pNsc->MacStats.txUnderrunCount;
	pAdpStats->collisions = pNsc->MacStats.txTotCollisionCount;

	pAdpStats->tx_bytes = pNsc->MacStats.txByteCount;
	pAdpStats->rx_bytes = pNsc->MacStats.rxByteCount;

	pAdpStats->rx_length_errors = 0;
	pAdpStats->rx_over_errors = 0;
	pAdpStats->rx_fifo_errors = 0;
	
	pAdpStats->tx_heartbeat_errors = 0;
	pAdpStats->tx_window_errors = 0;

 	return pAdpStats;
}


/******************************************************************************
*f*
*   Name:           init_module 
*
*   Description:    init_module initializes the driver module. This routine 
*                   is called to initialize the driver when the module in 
*                   installed into a running kernel via insmod.
*
*   Parameters:     None 
*
*   Return Value:   OK or ENODEV 
*
*f*
***************************************************************************/

int
init_module (void)
{

#ifdef _DUMP
	struct module 	*mlist = &__this_module;
	printk("module %s is loaded at %p\n",mlist->name,
		&((char *)mlist)[mlist->size_of_struct]);
#endif

	return DrvProbe (NULL);
}

/******************************************************************************
*f*
*   Name:           cleanup_module 
*
*   Description:    cleanup_module initializes the driver module. This routine 
*                   is called to initialize the driver when the module in 
*                   installed into a running kernel via insmod.
*
*   Parameters:     None 
*
*   Return Value:   None 
*
*f*
***************************************************************************/

void
cleanup_module (void)
{
	struct device *cur_dev;
	struct device *pDev;
	AdapterContext	*pNsc;

	/* Traverse through the list of device structures */
	cur_dev= DevList;
	while (cur_dev) 
	{
		/* Get the next device structure in the list */
		pDev = (struct device *) (((NsmContext *)cur_dev->priv)->next);

		if(((NsmContext *)cur_dev->priv)->status & NSM_OPEN)
			AdapterClose(cur_dev);
		pNsc = &(((NsmContext *)(cur_dev->priv))->AdpCtxt);

		#ifdef LINK_AGGR
		if(pNsc->nFlags)
			LacpUnInitialize(pNsc);
		#endif

		/* Unregister the instance of the network device, corresponding to this 
			device structure  */
		unregister_netdev (cur_dev);
		
#ifdef IOMAPPED_IO
		/* Release the IO region held by this instance of the device */
		release_region(cur_dev->base_addr, PCI_IO_SIZE);
#else
		iounmap(pNsc->RegAddr);
#endif

		kfree(((NsmContext *)cur_dev->priv)->pAdapterStats);
	
		kfree(cur_dev); 
	
		/* Update the cur_dev to point to the next device in the list */
		cur_dev = pDev;
	}
}

/******************************************************************************
*f*
*   Name:           NsmTxCompleted 
*
*   Description:    This is called by HSM to indicate that transmission is 
*                   complete for a sublist of HsmTCBs in the HsmTCB list, 
*                   which is indicated by sending the first and the TCB, next 
*                   to the last TCB in list.
*
*   Parameters:     pNsc:	Pointer to the AdapterContext structure
*
*                   TcbFirst:    Pointer to the first HsmTCB in the list for 
*                   which transmit completion has happened
*
*                   TcbLast:     Pointer to the HsmTCB which is the next from 
*                   the last TCB in the list for which transmit completion has 
*                   happened
* 
*
*   Return Value:   None 
*
*f*
***************************************************************************/

void NsmTxCompleted(AdapterContext *pNsc, HsmTCB *TcbFirst, HsmTCB *TcbLast)
{

	HsmTCB *tcb;

	tcb = TcbFirst;

	/* Free up the data buffer associated with the indicated HsmTCBs */  
	do
	{
		if((struct sk_buff *)tcb->pOsCtrlBlk) 
		{
			kfree_skb((struct sk_buff *)tcb->pOsCtrlBlk);
			tcb->pOsCtrlBlk = NULL;
		}
		tcb = tcb->pNext;
	}while(tcb != TcbLast);
}

/******************************************************************************
*f*
*   Name:           NsmRxPackets 
*
*   Description:    This is called by the HSM, to send the received packets in 
*                   a particular priority, over to the upper layers
*
*   Parameters:     pNsc:	Pointer to the AdapterContext structure
*
*                   RcbStart:	Pointer to the first HsmRCB in the list for 
*                   which receive completion has happened
*     
*                   RcbEnd:	Pointer to the HsmRCB which is the next from 
*                   the last RCB in the list for which receive completion has 
*                   happened
*
*                   Priority:    Priority queue on which the packet was obtained
*
*   Return Value:   None
*
*f*
***************************************************************************/

STATIC VOID 
NsmRxPackets (AdapterContext *pNsc, HsmRCB *RcbStart, HsmRCB *RcbEnd, 
	UCHAR Priority) 
{
	HsmRCB 			*rcb;
	HsmRCB 			*CurrRCB;
	HsmRCB 			*LastRCB;
	HsmRCB 			*TempRcb;
	struct sk_buff 	*NewSkb;
	struct sk_buff	*skb_new;
	struct sk_buff	*skb;
	UINT 			RemainingRCBs = 0;
	struct device 	*dev;
	UINT 			BufSizeReqd;

	/* Get the pointer to the NsmContext */
	NsmContext 		*pNsm = (NsmContext *)pNsc->pNsmContext; 

	/* Get the start of the list of HsmRCBs for which Rx complete has happened  
		*/
	rcb = RcbStart;

	dev = pNsm->dev;
#ifdef LINK_AGGR
	if(pNsc->nFlags & GEC_FLAG)
		dev = pNsm->dev_aggr;
#endif

	/* Traverse through the HsmRCB list and for each RCB from RcbStart to 
		RcbEnd */
	while ( rcb != RcbEnd )
	{

		/* check for jumbo frames by checking the More flag in HsmRCB, if it's a
			jumbo packet */
		if(rcb->More)
		{

			/* Get the start of the HsmRCB list */
			CurrRCB = rcb;

			/* Initialise the variable to update the size of the jumbo packet 
				received */
			BufSizeReqd = 0;

			/* For all the HsmRCBs into which this packet has spanned, do the 
				following */
			do
			{
				/* Calculate the size of the entire packet by adding the PktSize
					in the individual HsmRCBs, constituting the packet */
				BufSizeReqd += rcb->PktSize;

				/* Get the next HsmRCB in the RCB list */
				rcb = rcb->pNext;
			} while(rcb->More); 
	
			LastRCB = rcb;

			/* Update with the size of the last fragment of the packet */
			BufSizeReqd += rcb->PktSize;
			
			/* Get the next HsmRCB in the RCB list */
			rcb = rcb->pNext;
			/* Allocate a new sk_buff to hold the complete jumbo packet, thereby
				one having a data buffer of enough size */
			NewSkb = alloc_skb(BufSizeReqd + HEAD_SPACE, GFP_ATOMIC);

			if( NewSkb == NULL)
			{

#ifdef FAILURE_MESSAGES
				printk("The allocation of the sk_buff of size %d for the jumbo packet FAILED\n", BufSizeReqd);
#endif
				/* Update statisitics */
				pNsc->MacStats.rxErrorCount++;
				pNsc->MacStats.rxDroppedCount++;

				continue;	
			}
#ifdef PRIORITY_QUEUES			
			/* Set the priority field of the new sk_buff */
			NewSkb->priority = Priority;
#else
			NewSkb->priority = 0;
#endif

			/* Reserve some space at the beginning of the buffer */
			skb_reserve(NewSkb, HEAD_SPACE);
			
			/* If checksum offloading is supported ie if it is already done in 
				hardware then indicate it in the skbuff */

#ifdef CHECKSUM
			if (LastRCB->ExtSts & (TCPERR|UDPERR))
				NewSkb->ip_summed = CHECKSUM_NONE;
			else
				NewSkb->ip_summed = CHECKSUM_UNNECESSARY;
#else
			NewSkb->ip_summed = CHECKSUM_NONE;
#endif

			/* Traverse through the HsmRCB list to copy the entire jumbo 
				packet into the newly allocated sk_buff */
			while(CurrRCB != rcb)
 			{
				memcpy( NewSkb->tail, 
					((struct sk_buff *)CurrRCB->pOsCtrlBlk)->data, 
						CurrRCB->PktSize);		
				skb_put(NewSkb, CurrRCB->PktSize); 

				CurrRCB->Status = RCB_FREE;
				
				CurrRCB = CurrRCB->pNext;
			}

			NewSkb->dev = dev;

			NewSkb->protocol = eth_type_trans(NewSkb,dev);

			/* Send the new sk_buff up */
			netif_rx(NewSkb);
		}
		else
		{
			/* Through the HsmRCB, get the pointer to the OsCtrlBlk 
				( sk_buff ) */

			skb = (struct sk_buff *)rcb->pOsCtrlBlk;

			/* Clone this sk_buff */
			skb_new = skb_clone(skb, GFP_ATOMIC); 

			if( skb_new == NULL)
			{
				printk(KERN_ERR "NsmRxPackets : The cloning failed\n");
				TempRcb = rcb;

				/* Find out the number of packets remaining, which are now 
				going to be dropped*/
				while((TempRcb != RcbEnd) && TempRcb)
				{
					while(TempRcb->More) 
						TempRcb = TempRcb->pNext;
					RemainingRCBs++;
					TempRcb = TempRcb->pNext;
				}	
				
				/* Update statistics */
				pNsc->MacStats.rxErrorCount += RemainingRCBs; 
				pNsc->MacStats.rxDroppedCount += RemainingRCBs; 

				break;
			}
			
			rcb->Status = DUPED;

			skb_put(skb_new,rcb->PktSize);

			skb_new->dev = dev;
			skb_new->protocol = eth_type_trans(skb_new,dev);

			/* If checksum offloading is supported ie if it is already 
			   done in hardware then indicate it in the skbuff */
#ifdef CHECKSUM
			if (rcb->ExtSts & (TCPERR|UDPERR))
				skb_new->ip_summed = CHECKSUM_NONE;
			else
				skb_new->ip_summed = CHECKSUM_UNNECESSARY;
#else
			skb_new->ip_summed = CHECKSUM_NONE;			
#endif

#ifdef PRIORITY_QUEUES			
			/* Set the priority field of the new sk_buff */
			skb_new->priority = Priority;
#else
			skb_new->priority = 0;
#endif

			netif_rx(skb_new);  

			/* Get the next HsmRCB in the RCB list */
			rcb = rcb->pNext;
		}
	}

	/* Set the time for the last receive */
	pNsm->dev->last_rx = jiffies;

}

/******************************************************************************
*f*
*   Name:           NsmMallocContiguous 
*
*   Description:    This is called by HSM to allocate a contiguous chunk of 
*                   memory, the virtual & physical addresses of the start of 
*                   the chunk of memory, is set in the Addrstruct structure. 
*                   The byte alignment required is given by the Alignment 
*                   parameter.
*
*   Parameters:     pNsm: Pointer to the NsmContext    
*              
*                   pAddrstruct: Pointer to the Addrstruct structure, where the 
*                   virtual and physical addresses need to be put
*
*                   BitAlignment:   The alignment factor, to which the memory 
*                   allocated needs to be aligned
*
*                   Cached: Cached/Non-Cached memory
*
*   Return Value:   SUCCESS or FAILURE 
*
*f*
***************************************************************************/

UCHAR  NsmMallocContiguous(VOID *pNsm,AddrStruct *pAddrStruct, 
	USHORT BitAlignment, UINT Cached)
{

	VOID 	*pMem;

	UINT 	AlignedMem;

	USHORT 	Alignment = BitAlignment/8;
		
	if(in_interrupt())
		pMem = kmalloc (pAddrStruct->BufSize + Alignment -1, GFP_ATOMIC);
	else
		pMem = kmalloc (pAddrStruct->BufSize + Alignment -1, GFP_KERNEL);
	
	if(pMem == NULL)
	{
#ifdef FAILURE_MESSAGES
		printk("\nAllocation of contiguous memory failed");
#endif
		return FAILURE;
	}
	
	NsmZeroMemory(pMem, pAddrStruct->BufSize + Alignment - 1);	

	AlignedMem = ((UINT) pMem) + (Alignment -1);

	pAddrStruct->pCookie = pMem;

	AlignedMem &= ~(Alignment -1);

	pAddrStruct->VirtAddr = (UCHAR  *)AlignedMem;

#ifdef ASSERTION
	if(AlignedMem & (Alignment-1))
		NsmDbgMsg1("ASSERTION FAILED: In NsmMallocContiguous, the allocated memory is not %d byte aligned\n",Alignment);
#endif

	pAddrStruct->PhysAddr = (UCHAR *)virt_to_bus ((VOID *)AlignedMem);
	return SUCCESS;
	 
}

/******************************************************************************
*f*
*   Name:           NsmFreeContiguous
*
*   Description:    This is called by HSM to free a contiguous chunk of memory 
*
*   Parameters:     pNsm: Pointer to the NsmContext
*
*                   pAddrstruct: Pointer to the Addrstruct structure, where the 
*                   virtual and physical addresses are put
*
*                   Cached: Cached/Non-Cached memory
*
*   Return Value:   None
*
*f*
******************************************************************************/
			
void NsmFreeContiguous(VOID *pNsm, AddrStruct *pAddrStruct, UINT Cached)
{
	kfree(pAddrStruct->pCookie);
	pAddrStruct->VirtAddr = 0;
	pAddrStruct->PhysAddr = 0;
	pAddrStruct->pCookie = 0;
}

/******************************************************************************
*f*
*   Name:           NsmAllocatePacket
*
*   Description:    This is called by HSM to allocate an sk_buff with a data 
*                   buffer associated with it.
*
*   Parameters:     pNsm:    Pointer to the NsmContext structure
*
*                   pRcb:    Pointer to the HsmRCB, to which the allocated 
*                   packet is to be linked to
*
*                   pAddr: Pointer to the Addrstruct structure, where the 
*                   virtual and physical addresses are put
*
*   Return Value:   SUCCESS or FAILURE
*
*f*
******************************************************************************/

UCHAR  NsmAllocatePacket(VOID *pNsm, struct _HsmRCB *pRcb, AddrStruct *pAddr)
{

	struct sk_buff 	*skb;
	skb = alloc_skb(pAddr->BufSize, GFP_ATOMIC);
	
	if(skb == NULL)
	{
#ifdef FAILURE_MESSAGES
		printk("Allocation of the sk_buff failed\n");
#endif
		return FAILURE;
	}
	skb->dev = ((NsmContext *)pNsm)->dev;
	NsmZeroMemory(skb->data, pAddr->BufSize);	

	pAddr->pCookie = (VOID *)skb;
	pAddr->PhysAddr = (UCHAR *)virt_to_bus(skb->data);
	pRcb->pDataBuf = pAddr->VirtAddr = skb->data;

#ifdef ASSERTION
	if(((UINT)skb->data) & 7)
		NsmDbgMsg("ASSERTION FAILED : In NsmAllocatePacket, the allocated memory is not 8 byte aligned\n");
#endif

	pRcb->pOsCtrlBlk = (VOID *)skb;
	return SUCCESS;
}

/******************************************************************************
*f*
*   Name:           NsmDeAllocatePacket
*
*   Description:    This is called by HSM to deallocate the sk_buff structure 
*                   and consequently the data buffer associated with it, within
*                   an HsmRCB
*
*   Parameters:     pNsm: Pointer to the NsmContext
*
*					Addr: Pointer to the AddrStruct
*
*					pRcb:    Pointer to the HsmRCB, to which the allocated 
*                   packet is to be linked to
*
*   Return Value:   None
*
*f*
******************************************************************************/

void NsmDeAllocatePacket(VOID *pNsm, AddrStruct *Addr, HsmRCB *pRcb)
{
	kfree_skb(pRcb->pOsCtrlBlk);
	pRcb->pOsCtrlBlk = NULL;
	pRcb->pDataBuf = NULL;
}

/* Utility routines */

/******************************************************************************
*f*
*   Name:         CreatePacketstruct
*
*   Description:  This creates a packet struct and encapsulates the data in 
*                 the sk_buff in to the Pktstruct
*
*   Parameters:   skb: Pointer to the sk_buff, from which the data to be sent 
*                 down is taken to be encapsulated in to the Pktstruct data 
*                 structure. 
*    
*                 pPkt:  Pointer to the Pktstruct structure, into which the 
*                 data to be sent down is put
*
*   Return Value: None
*
*f*
******************************************************************************/

void 
CreatePacketStruct(struct sk_buff *skb, PktStruct *pPkt)
{
	/* Set the pointer to the physical memory for this fragment in the packet 
		structure */
	pPkt->PktArray[0].FragArray[0].pFrag = 
		(UCHAR *)virt_to_bus(skb->data);

	/* Update the length field of the fragment */
	pPkt->PktArray[0].FragArray[0].FragLen = skb->len; 

#ifdef STATISTICS
	pPkt->PktArray[0].TotalDataLength = skb->len;
#endif

	/* Put the number of fragments within the packet in the NumOfFrags field 
		within the packet struct */
	pPkt->PktArray[0].NumFrags = 1;	

	/* Make the pOsCtrlBlk point to the skb */
	pPkt->PktArray[0].pOSCtrlBlk = skb;	
	
	pPkt->PktArray[0].ChkSum = 0;	

	pPkt->PktCount = 1;

}

/******************************************************************************
*f*
*   Name:         CreatePacketInfo
*
*   Description:  This creates a packet info structure and encapsulates 
*                 the data in the sk_buff in to the PktInfo structure
*
*   Parameters:   skb: Pointer to the sk_buff, from which the data to be sent 
*                 down is taken to be encapsulated in to the PktInfo data 
*                 structure. 
*    
*                 pPkt:  Pointer to the PktInfo structure, into which the 
*                 data to be sent down is put
*
*   Return Value: None
*
*f*
******************************************************************************/

VOID 
CreatePacketInfo(struct sk_buff *skb, PktInfo *pPkt)
{

	/* Set the pointer to the physical memory for this fragment in the packet 
		structure */
	pPkt->FragArray[0].pFrag = 
		(UCHAR *)virt_to_phys(skb->data);

	/* Update the length field of the fragment */
	pPkt->FragArray[0].FragLen = skb->len; 

#ifdef STATISTICS
	pPkt->TotalDataLength = skb->len;
#endif

	/* Put the number of fragments within the packet in the NumOfFrags field 
		within the packet struct */
	pPkt->NumFrags = 1;	

	/* Make the pOsCtrlBlk point to the skb */
	pPkt->pOSCtrlBlk = skb;	
	pPkt->ChkSum = 0;	

}

/******************************************************************************
*f*
*   Name:         IsBrdCastOrMultiCast
*
*   Description:  This is used to check whether a particular address is a 
*                 broadcast or multicast address address
*
*   Parameters:   dev: Pointer to the device structure
*
*                 DestMac: Pointer to the destination MAC address
*
*   Return Value: SUCCESS or FAILURE
*
*f*
******************************************************************************/


UCHAR IsBrdCastOrMultiCast(struct device *dev, UCHAR *DestMac)
{
	struct dev_mc_list *mc_list;

	/* Check for broadcast address */
	if ((DestMac[0]==0xff) && (DestMac[1]==0xff) 
		&& (DestMac[2]==0xff) && (DestMac[3]==0xff) 
			&& (DestMac[4]==0xff) && (DestMac[5]==0xff))
		return SUCCESS;
	
	/* Check for multicast address */
	if(DestMac[0]==0x01)
	{

		mc_list = dev->mc_list;

		while(mc_list)
		{
			if(CMP_ETH_ADDR(mc_list->dmi_addr, (UCHAR *)DestMac, MAC_ADDR_LEN))
				return SUCCESS;
			
			mc_list = mc_list->next;
			
		}
		return FAILURE;
	}

	else
		return FAILURE;		    
}

/******************************************************************************
*f*
*   Name:         ChangeMtu
*
*   Description:  This changes the MTU size in the device structure to the new 
*                 value specified by NewMtu
*
*   Parameters:   dev:     Pointer to the device structure
*
*                 NewMtu:  The new MTU size which needs to be set in the device 
*                 structure
*
*   Return Value: EINVAL or OK
*
*f*
******************************************************************************/

STATIC int ChangeMtu(struct device *dev, int NewMtu)
{
	NsmContext	*pNsm = (NsmContext *)dev->priv;
	AdapterContext	*pNsc = &pNsm->AdpCtxt;
	USHORT	iff_up=0;
	USHORT	PktSize = NewMtu;

	/* Check for the accepatable range of MTU sizes */
	if((NewMtu < (MINIMUM_ETH_PACKET_SIZE - ETHER_HEADER_LENGTH))
		|| (NewMtu > (MAX_JUMBO_FRAME_SIZE - ETHER_HEADER_LENGTH)))
			return -EINVAL;

	if(PktSize % RXBUF_SIZE_ALIGN)
		PktSize += (RXBUF_SIZE_ALIGN -(PktSize % RXBUF_SIZE_ALIGN));

	if(pNsc->MaxPktSize != PktSize)
	{
		/* Change the RxBufSize to NewMtu and reinitialize the
		the device */

		iff_up = (dev->flags & IFF_UP);
		if(iff_up) 
		{
			/* Mark the interface down */
			dev->flags &= ~(IFF_UP);

			/* Stop the transmission through the interface*/
			netif_stop_queue(dev);

			MOD_DEC_USE_COUNT;
			del_timer(&pNsm->LedTimer);

			/* Sleep for 1000 milliseconds for any pending transmit to 
			complete before releasing the resource */
			NsmSleep(pNsc,1000);
		}
	
		if(pNsm->status & NSM_OPEN)
			AdapterClose(dev);
		pNsc->MaxPktSize = PktSize;
		if((AdapterOpen(dev)) != OK)
		{
			printk (KERN_ERR "%s: Not Enough Memory: For the interface %s to work, unload & load the driver\n",DRV_NAME,dev->name);
			if(dev->flags & IFF_RUNNING)
				dev->flags &= ~(IFF_RUNNING);
			return -ENOMEM;
		}

		if(iff_up)
		{
			/* Start the transmission through the interface */
			netif_start_queue(dev);
			/* Mark the interface Up */
			dev->flags |= IFF_UP; 
			MOD_INC_USE_COUNT;
			init_timer(&pNsm->LedTimer);
		}
	}

	/* Set the mtu field in the device structure to the NewMtu */
	dev->mtu = NewMtu;
	return (0);
}
		
/******************************************************************************
*f*
*   Name:         CMP_ETH_ADDR
*
*   Description:  This is used to compare two ethernet ( MAC ) addresses
*                 
*   Parameters:   SrcMac: Address of source 
*
*                 DstMac: Address of destination 
*                 
*                 Len: Number of bytes to be compared
*   
*	Return Value: 0 or 1 
*
*f*
******************************************************************************/

inline UCHAR CMP_ETH_ADDR(UCHAR *SrcMac,UCHAR *DestMac, UINT Len)
{
	UINT i = 0;
	while(i<Len)
	{
       	if(DestMac[i]!=SrcMac[i])
			break;	
		i++;
	}
	if(i == Len)
		return 0;
    else
        return 1;
}                                                                              

/******************************************************************************
*f*
*   Name:         NsmCreateLock
*
*   Description:  Allocates the data structure for a lock and intialises it
*
*   Parameters:   None
*
*   Return Value: Pointer to the LockId
*
*f*
******************************************************************************/

inline VOID *NsmCreateLock()
{
	VOID *LockId = NsmMalloc(sizeof(spinlock_t));
	
	if(LockId == NULL)
		return NULL;
	
	spin_lock_init((spinlock_t *)LockId);	
	return LockId;
}

/******************************************************************************
*f*
*   Name:         NsmAcquireLock
*
*   Description:  Acquires a lock which has been initialized early on
*
*   Parameters:   pAdapter: Pointer to the AdapterContext 
*
*                 LockId: Pointer to the data structure allocated for the lock 
*
*   Return Value: None
*
*f*
******************************************************************************/

inline VOID NsmAcquireLock(AdapterContext *pAdapter,VOID *LockId)
{
	spin_lock((spinlock_t *)LockId);
}

/******************************************************************************
*f*
*   Name:         NsmReleaseLock
*
*   Description:  Release a lock which has been acquired early on
*
*   Parameters:   pAdapter: Pointer to the AdapterContext 
*
*                 LockId: Pointer to the data structure allocated for the lock 
*
*   Return Value: None
*
*f*
******************************************************************************/

inline VOID NsmReleaseLock(AdapterContext *pAdapter, VOID *LockId)
{
	spin_unlock((spinlock_t *)LockId);
}

/******************************************************************************
*f*
*   Name:         NsmMalloc
*
*   Description:  Allocates kernel memory of the specified size
*
*   Parameters:   size: Size of the memory that needs to be allocated
*
*   Return Value: Pointer to the memory that is allocated
*
*f*
******************************************************************************/

inline VOID *NsmMalloc(UINT size)
{
	VOID *ptr;

	if(in_interrupt())
		ptr = kmalloc(size, GFP_ATOMIC);
	else
		ptr = kmalloc(size, GFP_KERNEL);

	if(ptr == NULL)
		return NULL;
	
	NsmZeroMemory(ptr,size);
	return ptr;
}

/******************************************************************************
*f*
*   Name:         PrintContents
*
*   Description:  Prints the contents of kernel memory locations in the range
*                 specified 
*
*   Parameters:   start: Start address 
*
*                 len: Length of the range 
*
*   Return Value: None
*
*f*
******************************************************************************/

inline VOID PrintContents(UCHAR *start, UINT len)
{
	UCHAR	*ptr = start;
	UINT 	i;
	printk("The contents of the location :\n");
	for(i=0;i<len;i++)
	{
		printk(": %x :"	,*ptr);			
		ptr++;
	}
}

UCHAR AdapterOpen (struct device *dev)
{
	UINT 			stat;
	NsmContext 		*pNsm;
	AdapterContext 	*pNsc;
	UINT 			dp_val;
	UINT			i;
	
#ifdef FT_DBG
	static UINT 	CallCount=0;
	printk("The DrvOpen call No : %d\n",CallCount++);
#endif

	/* Get the pointer to the allocated private data structure, NsmContext, 
	from the device structure */
	pNsm = (NsmContext *)dev->priv;

	/* Get the pointer to the AdapterContext structure from the NsmContext */
	pNsc = &pNsm->AdpCtxt;

	if (vlan)
		pNsc->VLANId = vlan & 0xFFF;

	for(i=0; i< MAX_PRI_QUEUE; i++)
	{
		pNsc->RxQueueSz[i] = MaxRxDesc;
		pNsc->TxQueueSz[i] = MaxTxDesc;
	}
	pNsc->TxDrth = 200;		
	pNsc->RxDrth = 0;
	pNsc->TxFlth = 2;
	pNsc->PauseCounterVal = 40000;
	pNsc->RxFFLO = 2;
	pNsc->RxFFHI = 8;
	pNsc->RxSTLO = 2;
	pNsc->RxSTHI = 8;

	/* Initialize fields which would indicate the capabilities of the hardware,
	 such as the, checksum offload */
	HsmInitContext(pNsc);
	if (vlan)
		pNsc->PhysCapabilities |= (VLAN_TAG_INSERTION_GEN_ON | VLAN_TAG_REMOVAL_ON);

	pNsc->PhysCapabilities &= ~DISCARD_VLAN_TAGGED_PACKETS; 

#ifdef NSCDEBUG
	printk(KERN_INFO "HsmInitContext called\n");
#endif

	/* Find out the capabilities of the OS and indicate whether the above 
	capabilties are supported by the OS. */

	if(MaxPrio > 8)
		MaxPrio = 8;
	
	pNsc->OsPrio = MaxPrio;
	pNsc->PhysCapabilities &= ~(SET_BIG_ENDIAN_MODE | TX_CHKSUM_OFFLOAD_ON_GEN |
			SET_RX_CHKSUM_OFFLOAD_ON );

	if(SetAutoNeg == 0)
	{
		pNsc->PhysCapabilities &= ~AUTO_NEG_ON;  
	}
	pNsc->UserMediaSpeed = MediaSpeed;
	pNsc->UserDuplexMode = Duplex;
	pNsc->NcBit = NCBit;

#ifdef CHECKSUM
	pNsc->PhysCapabilities |= (RX_IP_CHKSUM_DISCARD_ON | SET_RX_CHKSUM_OFFLOAD_ON);
#endif
	
	/* Allocate the Tx and Rx lists, based on the number of priority levels 
	supported. Call the HsmInitialise routine of the HSM, this will also setup 
	transmit & receive control. This will set up descriptor lists for each 
	level of priorities on both send and receive side. */

	if ( (stat = HsmInitialize(pNsc) ) == FAILURE )
	{
#ifdef FAILURE_MESSAGES
		printk (KERN_ERR "%s: HsmInitialize Failed \n", DRV_NAME);
#endif
		return -EAGAIN;
	}

#ifdef NSCDEBUG
	
	printk("The value of the tail %p 's CMDSTS : %x\n" , pNsc->PriQue[0].pRcbListTail->pRxDp, pNsc->PriQue[0].pRcbListTail->pRxDp->CmdSts);

#endif


#ifdef NSCDEBUG
	printk("The number of RX priorities : %d\n",pNsc->RxNumPrio);
#endif

	/* Now set the Mac address in the device structure */
	NsmCopy((char *)pNsc->PermMacAddr, (char *)dev->dev_addr, 
		MAC_ADDR_LEN); 

	del_timer(&pNsm->LedTimer); //any ghosts?
	init_timer(&pNsm->LedTimer);
	
	/* Install interrupt vector and register the interrupt service routine */
	if(request_irq (dev->irq, DrvIsr, SA_SHIRQ,
    	dev->name, dev) != OK)
	{
#ifdef FAILURE_MESSAGES
		printk (KERN_ERR "%s: request_irq Failed \n", DRV_NAME);
#endif
		HsmUnInitialize(pNsc);
		return -EAGAIN;
	}

	/* Reset the adapter  */
	if( ( stat =HsmReset(pNsc, 0) ) == FAILURE)
	{
#ifdef FAILURE_MESSAGES
		printk (KERN_ERR "%s: Resetting the adapter, failed \n", 
			DRV_NAME);
#endif
		free_irq (dev->irq, dev);
		HsmUnInitialize(pNsc);
		return -EAGAIN;
	} 

	/* Call HsmOpen call to open the adapter */
	if( ( stat =HsmOpen(pNsc, 1) ) == FAILURE)
	{
#ifdef FAILURE_MESSAGES
		printk (KERN_ERR "%s: Opening the adapter, failed \n", 
			DRV_NAME);
#endif
		free_irq (dev->irq, dev);
		del_timer(&pNsm->LedTimer);
		HsmUnInitialize(pNsc);
		return -EAGAIN;
	}
	
	HsmRxFilter(pNsc, (ACCEPT_PERFECT_MATCH_ON | ACCEPT_ALL_BROADCAST_ON));

#if 0
	/* Mark the start flag in the device structure */
	dev->start = DEV_STARTED;

	dev->tbusy = 0;
	dev->interrupt = 0;

#endif 

	/* Set the transmit queue length in the device structure */
	dev->tx_queue_len = (MaxTxDesc*MaxPrio);


	NsmRegRead32(pNsm, (pNsc->RegAddr + RXDP), &dp_val);

#ifdef NSCDEBUG			
	NsmDbgMsg1("RXDP's initial Val : %x\n", dp_val);
#endif

	pNsm->status |= NSM_OPEN;
    return OK;
}


void AdapterClose (struct device *dev)
{
	NsmContext 		*pNsm;
	AdapterContext 	*pNsc;

#ifdef FT_DBG
	static UINT 	CallCount=0;
	printk("The DrvClose call No : %d",CallCount++);
#endif

#if 0
	dev->start = 0;
	dev->tbusy = 1;
#endif

	/* Get the pointer to the NsmContext and thereby to the AdapterContext, 
	from the device structure */
	pNsm = (NsmContext *)dev->priv;
	pNsc   = &(pNsm->AdpCtxt);

#ifdef NSCDEBUG
	printk(KERN_INFO "DrvClose called\n");
#endif

	/* Stop Tx & Rx by calling HsmClose */
	HsmClose(pNsc); 

	/* Uninstall the interrupt vector, and deregister the interrupt service 
	routine */
	free_irq (dev->irq, dev);

	/* Sleep for any interrupt which arrived just before the HsmClose and for 
	which HsmIsr has not yet read the ISR register */
	do
	{
		NsmSleep(pNsc,200);
	}while(pNsc->interruptStatus || (pNsm->LedTimer.expires > jiffies) );

#ifdef LINK_AGGR
	if(pNsc->nFlags)
		LacpDisablePort(pNsc);
#endif
	del_timer(&pNsm->LedTimer);

	/* Free the Tx and Rx lists by calling HsmUnInitialize */
	HsmUnInitialize(pNsc);  
	pNsm->status &= ~NSM_OPEN;
}

void 
NsmIndicateLinkStatus(AdapterContext *pAdapter) 
{
	return;
}

void
NsmStartLedWithTimer(NsmContext *pNsm)
{
	pNsm->LedTimerSetAt = jiffies;

	if(pNsm->LedTimer.function == NULL)
	{
		/* Turn On Activity LED */
		HsmTbiActivityLedOn(&(pNsm->AdpCtxt));

		pNsm->LedTimer.expires = (pNsm->LedTimerSetAt) + TBI_LED_TIMER_DELAY;
		pNsm->LedTimer.data = (unsigned long) &(pNsm->AdpCtxt);
		pNsm->LedTimer.function = LedTimerFunc;
		add_timer(&pNsm->LedTimer);
	}
}


void
LedTimerFunc(unsigned long data)
{
	AdapterContext	*pAdapter = (AdapterContext *) data;
	NsmContext	*pNsm = pAdapter->pNsmContext;
	unsigned	int	Elapsed;

	Elapsed = jiffies - pNsm->LedTimerSetAt;

	if( Elapsed >= TBI_LED_TIMER_DELAY)
	{
		pNsm->LedTimer.function = NULL;
		/* Turn Off Activity LED */
		HsmTbiActivityLedOff(pAdapter);
	}
	else
	{
		pNsm->LedTimer.expires = jiffies + TBI_LED_TIMER_DELAY;
		add_timer(&pNsm->LedTimer);
	}
}
